import os
import subprocess
import pandas
import matplotlib.pyplot as plt

""" 
Exploit genumfp command line tool for building histograms
"""


def compute_histogram(values_filepath: str,
                      histogram_filepath: str):
    """  Compute histogram using the G-Enum-fp method.
    This method is a wrapper on the command line tool khisto.

    @param values_filepath: path of file containing input data
    @param histogram_filepath: path of file containing the resulting histogram

    @return: success indicator
    @rtype bool
    """
    # Create output directory
    output_directory = os.path.dirname(histogram_filepath)
    if not os.path.isdir(output_directory):
        os.makedirs(output_directory)
    # Compute histogram from command line
    tool_file_name = os.path.join('..', 'khisto' + ('.exe' if os.name == 'nt' else ''))
    params = [tool_file_name, values_filepath, histogram_filepath]
    return_code = subprocess.call(params)
    return return_code == 0


def plot_histogram(histogram_filepath: str,
                   figure_filepath: str,
                   log_xscale: bool = False,
                   log_yscale: bool = False):
    """  Plot a histogram built using the G-Enum-fp method.

    @param histogram_filepath: path of histogram file build using the G-Enum-fp method
    @param figure_filepath: path of figure file to plot
    @param log_xscale: if True, use a log_10 scale for x_axis
    @param log_yscale: if True, use a log_10 scale for y_axis
    @rtype None
    """
    # Create output directory
    output_directory = os.path.dirname(figure_filepath)
    if not os.path.isdir(output_directory):
        os.makedirs(output_directory)
    # Extract bounds and frequencies from histogram file
    histogram_model = pandas.read_csv(histogram_filepath)
    lower_bounds = histogram_model['LowerBound'].tolist()
    upper_bounds = histogram_model['UpperBound'].tolist()
    frequencies = histogram_model['Frequency'].tolist()
    # Plot the histogram
    plt.clf()
    axis = plt.gca()
    if log_xscale:
        axis.set_xscale('log', basex=10)
    if log_yscale:
        axis.set_yscale('log', basey=10)
    plt.hist(lower_bounds, lower_bounds + upper_bounds[-1:], weights = frequencies, density=True,
             linewidth=0.5, edgecolor='black')
    plt.savefig(figure_filepath)

def compute_and_plot_histograms():
    """ Compute and plot histogram for each data set in data directory
    """
    for name in ['gaussian', 'levy', 'adult_age', 'lunar_crater_radius']:
        values_filepath = os.path.join('data', name + '.txt')
        # Compute histogram
        histogram_filepath = os.path.join('histograms', name + '_histogram.csv')
        compute_histogram(values_filepath, histogram_filepath)
        # Plot histogram, using log scale if necessary
        figure_filepath = os.path.join('histograms', name + '_histogram.png')
        log_scale = name in ['levy', 'lunar_crater_radius']
        plot_histogram(histogram_filepath, figure_filepath, log_xscale=log_scale, log_yscale=log_scale)

# Compute and plot histograms for all datasets
compute_and_plot_histograms()
