import os
import subprocess
import pandas
import matplotlib.pyplot as plt

""" 
Exploit khisto command line tool for exploratoy analysis using histograms
"""


def compute_histogram_series(values_filepath: str,
                             histogram_filepath: str):
    """  Compute a list of histograms for exploratory analysis using the G-Enum-fp method
    This method is a wrapper on the command line tool khisto.

    The parameter histogram_filepath is used as a base name to build the series of histograms
     with '.1', '.2' ... in the suffix, plus the file with synthetic indicators per histogram
     with '.series' in the suffix.

    @param values_filepath: path of file containing input data
    @param histogram_filepath: path of file containing the resulting histogram

    @return: success indicator and name of the file describing the series of histograms
    @rtype bool, str
    """
    # Create output directory
    output_directory = os.path.dirname(histogram_filepath)
    if not os.path.isdir(output_directory):
        os.makedirs(output_directory)
    # Build the name of the file describing the series of histograms
    filepath, file_extension = os.path.splitext(histogram_filepath)
    histogram_series_filepath = filepath + '.series' + file_extension
    # Compute histogram from command line, plus the list of histograms for exploratory analysis
    tool_file_name = os.path.join('..', 'khisto' + ('.exe' if os.name == 'nt' else ''))
    params = [tool_file_name, '-e', values_filepath, histogram_filepath]
    return_code = subprocess.call(params)
    return return_code == 0, histogram_series_filepath


def plot_histogram(histogram_filepath: str,
                   figure_filepath: str,
                   log_xscale: bool = False,
                   log_yscale: bool = False):
    """  Plot a histogram built using the G-Enum-fp method.

    @param histogram_filepath: path of file containing the resulting histogram
    @param figure_filepath: path of file containing figure to plot
    @param log_xscale: if True, use a log_10 scale for x_axis
    @param log_yscale: if True, use a log_10 scale for y_axis
    @rtype None
    """
    # Extract bounds and frequencies from histogram file
    histogram_model = pandas.read_csv(histogram_filepath)
    lower_bounds = histogram_model['LowerBound'].tolist()
    upper_bounds = histogram_model['UpperBound'].tolist()
    frequencies = histogram_model['Frequency'].tolist()
    # Plot the histogram
    plt.clf()
    axis = plt.gca()
    if log_xscale:
        axis.set_xscale('log', basex=10)
    if log_yscale:
        axis.set_yscale('log', basey=10)
    plt.hist(lower_bounds, lower_bounds + upper_bounds[-1:], weights = frequencies, density=True,
             linewidth=0.5, edgecolor='black')
    plt.savefig(figure_filepath)


def plot_histogram_series(histogram_series_filepath: str,
                          figure_directory: str,
                          log_xscale: bool = False,
                          log_yscale: bool = False):
    """  Plot a set of figures for exploratory analysis using histograms

    The method outputs a plot per histogram in the series, plus a synthetic plot with indicator curves
    per interpretable histogram.
    All the plots are stored in figure_directory, with names based of input model files and extension '.png'

    @param histogram_filepath: path of histogram series file build using the G-Enum-fp method
    @param figure_directory: directory where all figures to plot are output
    @param log_xscale: if True, use a log_10 scale for x_axis
    @param log_yscale: if True, use a log_10 scale for y_axis
    @rtype None
    """
    # Create output directory
    if not os.path.isdir(figure_directory):
        os.makedirs(figure_directory)
    # Extract information from from histogram series file
    histogram_series_information = pandas.read_csv(histogram_series_filepath)
    # Compute number of interpretable histograms, according to whether the last one is raw
    histogram_number = len(histogram_series_information.index)
    interpretable_histogram_number = histogram_number
    if histogram_series_information.iloc[histogram_number-1]['Raw'] == 1:
        interpretable_histogram_number -= 1
    # Plot the histograms of the series
    for index in range(histogram_number):
        histogram_filepath = histogram_series_information.iloc[index]['FileName']
        file_prefix, file_extension = os.path.splitext(os.path.basename(histogram_filepath))
        figure_filepath = os.path.join(figure_directory, file_prefix + ".png")
        # Plot the last one using a y log scale if it is not interpretable
        plot_histogram(histogram_filepath, figure_filepath,
                       log_xscale=log_xscale,
                       log_yscale=log_yscale or index == interpretable_histogram_number)
    # Plot synthetic interval stats per granularity, using a first axis
    plt.clf()
    figure, axis1 = plt.subplots()
    axis1.set_xlabel('granularity')
    axis1.set_ylabel('intervals')
    all_stats = ['IntervalNumber', 'PeakIntervalNumber', 'SpikeIntervalNumber', 'EmptyIntervalNumber']
    labels = ['all', 'peak', 'spike', 'empty']
    colors = ['black', 'darkgreen', 'red', 'blue']
    markers = ['s', '^', '2', '.']
    x = histogram_series_information['Granularity'].tolist()[:interpretable_histogram_number]
    for i, stats in enumerate(all_stats):
        y = histogram_series_information[stats].tolist()[:interpretable_histogram_number]
        axis1.plot(x, y, marker=markers[i], color=colors[i], label = labels[i])
    axis1.legend()
    # Plot information rate, using a second axes that shares the same x-axis
    axis2 = axis1.twinx()
    axis2_color = 'darkgray'
    axis2.set_ylabel('information rate', color=axis2_color)
    y = histogram_series_information['InformationRate'].tolist()[:interpretable_histogram_number]
    axis2.plot(x, y, color = axis2_color, marker='o')
    axis2.tick_params(axis='y', labelcolor=axis2_color)
    # Save figure
    figure.tight_layout()
    file_prefix, file_extension = os.path.splitext(os.path.basename(histogram_series_filepath))
    histogram_series_figure_filepath = os.path.join(figure_directory, file_prefix + ".png")
    plt.savefig(histogram_series_figure_filepath)

def histogram_exploratory_analysis():
    """ Compute and plot histogram for each data set in data directory
    """
    for name in ['gaussian', 'levy', 'adult_age', 'lunar_crater_radius']:
        values_filepath = os.path.join('data', name + '.txt')
        # Compute histogram
        histogram_filepath = os.path.join('exploratory_analysis', name, name + '_histogram.csv')
        ok, histogram_series_filepath = compute_histogram_series(values_filepath, histogram_filepath)
        if ok:
            figure_filepath = os.path.join('exploratory_analysis', name, name + '_histogram.png')
            figure_directory = os.path.join('exploratory_analysis', name)
            log_scale = name in ['levy', 'lunar_crater_radius']
            plot_histogram(histogram_filepath, figure_filepath, log_xscale=log_scale, log_yscale=log_scale)
            plot_histogram_series(histogram_series_filepath, figure_directory, log_xscale=log_scale, log_yscale=log_scale)

# Exploratory analysis for all datasets
histogram_exploratory_analysis()
